@@ -97,6 +97,7 @@
 
 #include "nsXBLPrototypeBinding.h"
 #include "nsXBLBinding.h"
+#include "nsBindingManager.h"
 #include "nsIPrincipal.h"
 #include "nsIScriptSecurityManager.h"
 #include "nsIEventListenerManager.h"
@@ -104,7 +105,7 @@
 
 #include "prprf.h"
 
-nsresult NS_DOMClassInfo_PreserveWrapper(nsIXPConnectWrappedNative *aWrapper);
+nsresult NS_DOMClassInfo_PreserveNodeWrapper(nsIXPConnectWrappedNative *aWrapper);
 
 // Helper classes
 
@@ -322,6 +323,7 @@ BuildContentLists(nsHashKey* aKey, void* aData, void* aClosure)
     nsCOMPtr<nsIContent> child(do_QueryInterface(node));
     if (((PRInt32)i) == currIndex) {
       // Add the currPoint to the supports array.
+      NS_IF_ADDREF(currPoint);
       contentList->AppendElement(currPoint);
 
       // Get the next real insertion point and update our currIndex.
@@ -337,15 +339,20 @@ BuildContentLists(nsHashKey* aKey, void* aData, void* aClosure)
     
     if (!pseudoPoint) {
       pseudoPoint = new nsXBLInsertionPoint(parent, (PRUint32) -1, nsnull);
-      contentList->AppendElement(pseudoPoint);
+      if (pseudoPoint) {
+        NS_ADDREF(pseudoPoint);
+        contentList->AppendElement(pseudoPoint);
+      }
+    }
+    if (pseudoPoint) {
+      pseudoPoint->AddChild(child);
     }
-
-    pseudoPoint->AddChild(child);
   }
 
   // Add in all the remaining insertion points.
   for ( ; j < count; j++) {
     currPoint = NS_STATIC_CAST(nsXBLInsertionPoint*, arr->ElementAt(j));
+    NS_IF_ADDREF(currPoint);
     contentList->AppendElement(currPoint);
   }
   
@@ -1164,7 +1171,7 @@ nsXBLBinding::InitClass(const nsCString& aClassName,
       do_QueryInterface(wrapper);
 
     if (native_wrapper) {
-      NS_DOMClassInfo_PreserveWrapper(native_wrapper);
+      NS_DOMClassInfo_PreserveNodeWrapper(native_wrapper);
     }
   }
 
@@ -1233,13 +1240,35 @@ nsXBLBinding::AllowScripts()
 
   PRBool canExecute;
   nsresult rv = mgr->CanExecuteScripts(cx, principal, &canExecute);
-  return NS_SUCCEEDED(rv) && canExecute;
+  if (NS_FAILED(rv) || !canExecute) {
+    return PR_FALSE;
+  }
+
+  // Now one last check: make sure that we're not allowing a privilege
+  // escalation here.
+  PRBool haveCert;
+  nsIPrincipal* targetDocPrincipal = doc->GetPrincipal();
+  if (!targetDocPrincipal) {
+    return PR_FALSE;
+  }
+  targetDocPrincipal->GetHasCertificate(&haveCert);
+  if (!haveCert) {
+    return PR_TRUE;
+  }
+
+  PRBool subsumes;
+  rv = principal->Subsumes(targetDocPrincipal, &subsumes);
+  return NS_SUCCEEDED(rv) && subsumes;
 }
 
 PR_STATIC_CALLBACK(PRBool)
 DeleteVoidArray(nsHashKey* aKey, void* aData, void* aClosure)
 {
-  delete NS_STATIC_CAST(nsVoidArray*, aData);
+  nsVoidArray* array = NS_STATIC_CAST(nsVoidArray*, aData);
+  if (array) {
+    array->EnumerateForwards(ReleaseInsertionPoint, nsnull);
+    delete array;
+  }
   return PR_TRUE;
 }
 
